# SPDX-License-Identifier: MIT

import redis as Redis

from drain3.persistence_handler import PersistenceHandler


class RedisPersistence(PersistenceHandler):
    def __init__(self, redis_key, redis_host='localhost', redis_port=6379, redis_db=0, redis_pass=None, is_ssl=False, redis=None):
        self.redis_key = redis_key

        if isinstance(redis, Redis):
            self.r = redis
        else:
            self.r = Redis(host=redis_host,
                                port=redis_port,
                                db=redis_db,
                                password=redis_pass,
                                ssl=is_ssl)

    def save_state(self, state):
        self.r.set(self.redis_key, state)

    def load_state(self):
        return self.r.get(self.redis_key)


# persistance = RedisPersistence(
#     redis= Redis(
#         host='localhost',
#         port=6379,
#         db=0,
#         password=None,
#         ssl=False
#     ),
#     redis_key='drain3'
# )


# SPDX-License-Identifier: MIT

from drain3.persistence_handler import PersistenceHandler


class MemoryBufferPersistence(PersistenceHandler):
    def __init__(self):
        self.state = None

    def save_state(self, state):
        self.state = state

    def load_state(self):
        return self.state


# SPDX-License-Identifier: MIT

import os
import pathlib

from drain3.persistence_handler import PersistenceHandler


class FilePersistence(PersistenceHandler):
    def __init__(self, file_path):
        self.file_path = file_path

    def save_state(self, state):
        pathlib.Path(self.file_path).write_bytes(state)

    def load_state(self):
        if not os.path.exists(self.file_path):
            return None

        return pathlib.Path(self.file_path).read_bytes()



# SPDX-License-Identifier: MIT

import kafka

from drain3.persistence_handler import PersistenceHandler


class KafkaPersistence(PersistenceHandler):

    def __init__(self, topic, snapshot_poll_timeout_sec=60, **kafka_client_options):
        self.topic = topic
        self.kafka_client_options = kafka_client_options
        self.producer = kafka.KafkaProducer(**self.kafka_client_options)
        self.snapshot_poll_timeout_sec = snapshot_poll_timeout_sec

    def save_state(self, state):
        self.producer.send(self.topic, value=state)

    def load_state(self):
        consumer = kafka.KafkaConsumer(**self.kafka_client_options)
        partition = kafka.TopicPartition(self.topic, 0)
        consumer.assign([partition])
        end_offsets = consumer.end_offsets([partition])
        end_offset = list(end_offsets.values())[0]
        if end_offset > 0:
            consumer.seek(partition, end_offset - 1)
            snapshot_poll_timeout_ms = self.snapshot_poll_timeout_sec * 1000
            records = consumer.poll(snapshot_poll_timeout_ms)
            if not records:
                raise RuntimeError(f"No message received from Kafka during restore even though end_offset>0")
            last_msg = records[partition][0]
            state = last_msg.value
        else:
            state = None

        consumer.close()
        return state



# SPDX-License-Identifier: MIT

import ast
import configparser
import json
import logging

from drain3.masking import MaskingInstruction

logger = logging.getLogger(__name__)


class TemplateMinerConfig:
    def __init__(self):
        self.engine = "Drain"
        self.profiling_enabled = False
        self.profiling_report_sec = 60
        self.snapshot_interval_minutes = 5
        self.snapshot_compress_state = True
        self.drain_extra_delimiters = []
        self.drain_sim_th = 0.4
        self.drain_depth = 4
        self.drain_max_children = 100
        self.drain_max_clusters = None
        self.masking_instructions = []
        self.mask_prefix = "<"
        self.mask_suffix = ">"
        self.parameter_extraction_cache_capacity = 3000
        self.parametrize_numeric_tokens = True

    def load(self, config_filename: str):
        parser = configparser.ConfigParser()
        read_files = parser.read(config_filename)
        if len(read_files) == 0:
            logger.warning(f"config file not found: {config_filename}")

        section_profiling = 'PROFILING'
        section_snapshot = 'SNAPSHOT'
        section_drain = 'DRAIN'
        section_masking = 'MASKING'

        self.engine = parser.get(section_drain, 'engine', fallback=self.engine)

        self.profiling_enabled = parser.getboolean(section_profiling, 'enabled',
                                                   fallback=self.profiling_enabled)
        self.profiling_report_sec = parser.getint(section_profiling, 'report_sec',
                                                  fallback=self.profiling_report_sec)

        self.snapshot_interval_minutes = parser.getint(section_snapshot, 'snapshot_interval_minutes',
                                                       fallback=self.snapshot_interval_minutes)
        self.snapshot_compress_state = parser.getboolean(section_snapshot, 'compress_state',
                                                         fallback=self.snapshot_compress_state)

        drain_extra_delimiters_str = parser.get(section_drain, 'extra_delimiters',
                                                fallback=str(self.drain_extra_delimiters))
        self.drain_extra_delimiters = ast.literal_eval(drain_extra_delimiters_str)

        self.drain_sim_th = parser.getfloat(section_drain, 'sim_th',
                                            fallback=self.drain_sim_th)
        self.drain_depth = parser.getint(section_drain, 'depth',
                                         fallback=self.drain_depth)
        self.drain_max_children = parser.getint(section_drain, 'max_children',
                                                fallback=self.drain_max_children)
        self.drain_max_clusters = parser.getint(section_drain, 'max_clusters',
                                                fallback=self.drain_max_clusters)
        self.parametrize_numeric_tokens = parser.getboolean(section_drain, 'parametrize_numeric_tokens',
                                                            fallback=self.parametrize_numeric_tokens)

        masking_instructions_str = parser.get(section_masking, 'masking',
                                              fallback=str(self.masking_instructions))
        self.mask_prefix = parser.get(section_masking, 'mask_prefix', fallback=self.mask_prefix)
        self.mask_suffix = parser.get(section_masking, 'mask_suffix', fallback=self.mask_suffix)
        self.parameter_extraction_cache_capacity = parser.get(section_masking, 'parameter_extraction_cache_capacity',
                                                              fallback=self.parameter_extraction_cache_capacity)

        masking_instructions = []
        masking_list = json.loads(masking_instructions_str)
        for mi in masking_list:
            instruction = MaskingInstruction(mi['regex_pattern'], mi['mask_with'])
            masking_instructions.append(instruction)
        self.masking_instructions = masking_instructions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drain3']

package_data = \
{'': ['*']}

install_requires = \
['cachetools', 'jsonpickle']

extras_require = \
{'all': ['redis', 'kafka-python'],
 'kafka': ['kafka-python'],
 'redis': ['redis']}

setup_kwargs = {
    'name': 'drain3-ankcorn',
    'version': '0.9.12',
    'description': 'Persistent & streaming log template miner',
    'long_description': '# Drain3\n\n## Important Update\n\nDrain3 was moved to the `logpai` GitHub organization (which is also the home for the original Drain implementation). We always welcome more contributors and maintainers to join us and push the project forward. We welcome more contributions and variants of implementations if you find practical enhancements to the algorithm in production scenarios.\n\n## Introduction\n\nDrain3 is an online log template miner that can extract templates (clusters) from a stream of log messages in a timely\nmanner. It employs a parse tree with fixed depth to guide the log group search process, which effectively avoids\nconstructing a very deep and unbalanced tree.\n\nDrain3 continuously learns on-the-fly and extracts log templates from raw log entries.\n\n#### Example:\n\nFor the input:\n\n```\nconnected to 10.0.0.1\nconnected to 192.168.0.1\nHex number 0xDEADBEAF\nuser davidoh logged in\nuser eranr logged in\n```\n\nDrain3 extracts the following templates:\n\n```\nID=1     : size=2         : connected to <:IP:>\nID=2     : size=1         : Hex number <:HEX:>\nID=3     : size=2         : user <:*:> logged in\n```\n\nFull sample program output:\n\n```\nStarting Drain3 template miner\nChecking for saved state\nSaved state not found\nDrain3 started with \'FILE\' persistence\nStarting training mode. Reading from std-in (\'q\' to finish)\n> connected to 10.0.0.1\nSaving state of 1 clusters with 1 messages, 528 bytes, reason: cluster_created (1)\n{"change_type": "cluster_created", "cluster_id": 1, "cluster_size": 1, "template_mined": "connected to <:IP:>", "cluster_count": 1}\nParameters: [ExtractedParameter(value=\'10.0.0.1\', mask_name=\'IP\')]\n> connected to 192.168.0.1\n{"change_type": "none", "cluster_id": 1, "cluster_size": 2, "template_mined": "connected to <:IP:>", "cluster_count": 1}\nParameters: [ExtractedParameter(value=\'192.168.0.1\', mask_name=\'IP\')]\n> Hex number 0xDEADBEAF\nSaving state of 2 clusters with 3 messages, 584 bytes, reason: cluster_created (2)\n{"change_type": "cluster_created", "cluster_id": 2, "cluster_size": 1, "template_mined": "Hex number <:HEX:>", "cluster_count": 2}\nParameters: [ExtractedParameter(value=\'0xDEADBEAF\', mask_name=\'HEX\')]\n> user davidoh logged in\nSaving state of 3 clusters with 4 messages, 648 bytes, reason: cluster_created (3)\n{"change_type": "cluster_created", "cluster_id": 3, "cluster_size": 1, "template_mined": "user davidoh logged in", "cluster_count": 3}\nParameters: []\n> user eranr logged in\nSaving state of 3 clusters with 5 messages, 644 bytes, reason: cluster_template_changed (3)\n{"change_type": "cluster_template_changed", "cluster_id": 3, "cluster_size": 2, "template_mined": "user <:*:> logged in", "cluster_count": 3}\nParameters: [ExtractedParameter(value=\'eranr\', mask_name=\'*\')]\n> q\nTraining done. Mined clusters:\nID=1     : size=2         : connected to <:IP:>\nID=2     : size=1         : Hex number <:HEX:>\nID=3     : size=2         : user <:*:> logged in\n```\n\nThis project is an upgrade of the original [Drain](https://github.com/logpai/logparser/blob/master/logparser/Drain)\nproject by LogPAI from Python 2.7 to Python 3.6 or later with additional features and bug-fixes.\n\nRead more information about Drain from the following paper:\n\n- Pinjia He, Jieming Zhu, Zibin Zheng, and Michael R.\n  Lyu. [Drain: An Online Log Parsing Approach with Fixed Depth Tree](http://jiemingzhu.github.io/pub/pjhe_icws2017.pdf),\n  Proceedings of the 24th International Conference on Web Services (ICWS), 2017.\n\nA Drain3 use case is presented in this blog\npost: [Use open source Drain3 log-template mining project to monitor for network outages](https://developer.ibm.com/blogs/how-mining-log-templates-can-help-ai-ops-in-cloud-scale-data-centers)\n.\n\n#### New features\n\n- [**Persistence**](#persistence). Save and load Drain state into an [Apache Kafka](https://kafka.apache.org)\n  topic, [Redis](https://redis.io/) or a file.\n- **Streaming**. Support feeding Drain with messages one-be-one.\n- [**Masking**](#masking). Replace some message parts (e.g numbers, IPs, emails) with wildcards. This improves the\n  accuracy of template mining.\n- [**Packaging**](#installation). As a pip package.\n- [**Configuration**](#configuration). Support for configuring Drain3 using an `.ini` file or a configuration object. \n- [**Memory efficiency**](#memory-efficiency). Decrease the memory footprint of internal data structures and introduce\n  cache to control max memory consumed (thanks to @StanislawSwierc)\n- [**Inference mode**](#training-vs-inference-modes). In case you want to separate training and inference phase, Drain3\n  provides a function for *fast* matching against already-learned clusters (templates) only, without the usage of\n  regular expressions.\n- [**Parameter extraction**](#parameter-extraction). Accurate extraction of the variable parts from a log message as an\n  ordered list, based on its mined template and the defined masking instructions (thanks to @Impelon).\n\n#### Expected Input and Output\n\nAlthough Drain3 can be ingested with full raw log message, template mining accuracy can be improved if you feed it with\nonly the unstructured free-text portion of log messages, by first removing structured parts like timestamp, hostname.\nseverity, etc.\n\nThe output is a dictionary with the following fields:\n\n- `change_type` - indicates either if a new template was identified, an existing template was changed or message added\n  to an existing cluster.\n- `cluster_id` - Sequential ID of the cluster that the log belongs to.\n- `cluster_size`- The size (message count) of the cluster that the log belongs to.\n- `cluster_count` - Count clusters seen so far.\n- `template_mined`- the last template of above cluster_id.\n\n## Configuration\n\nDrain3 is configured using [configparser](https://docs.python.org/3.4/library/configparser.html). By default, config\nfilename is `drain3.ini` in working directory. It can also be configured passing\na [TemplateMinerConfig](drain3/template_miner_config.py) object to the [TemplateMiner](drain3/template_miner.py)\nconstructor.\n\nPrimary configuration parameters:\n\n- `[DRAIN]/sim_th` - similarity threshold. if percentage of similar tokens for a log message is below this number, a new\n  log cluster will be created (default 0.4)\n- `[DRAIN]/depth` - max depth levels of log clusters. Minimum is 3. (default 4)\n- `[DRAIN]/max_children` - max number of children of an internal node (default 100)\n- `[DRAIN]/max_clusters` - max number of tracked clusters (unlimited by default). When this number is reached, model\n  starts replacing old clusters with a new ones according to the LRU cache eviction policy.\n- `[DRAIN]/extra_delimiters` - delimiters to apply when splitting log message into words (in addition to whitespace) (\n  default none). Format is a Python list e.g. `[\'_\', \':\']`.\n- `[MASKING]/masking` - parameters masking - in json format (default "")\n- `[MASKING]/mask_prefix` & `[MASKING]/mask_suffix` - the wrapping of identified parameters in templates. By default, it\n  is `<` and `>` respectively.\n- `[SNAPSHOT]/snapshot_interval_minutes` - time interval for new snapshots (default 1)\n- `[SNAPSHOT]/compress_state` - whether to compress the state before saving it. This can be useful when using Kafka\n  persistence.\n\n## Masking\n\nThis feature allows masking of specific variable parts in log message with keywords, prior to passing to Drain. A\nwell-defined masking can improve template mining accuracy.\n\nTemplate parameters that do not match any custom mask in the preliminary masking phase are replaced with `<*>` by Drain\ncore.\n\nUse a list of regular expressions in the configuration file with the format `{\'regex_pattern\', \'mask_with\'}` to set\ncustom masking.\n\nFor example, following masking instructions in `drain3.ini` will mask IP addresses and integers:\n\n```\n[MASKING]\nmasking = [\n          {"regex_pattern":"((?<=[^A-Za-z0-9])|^)(\\\\d{1,3}\\\\.\\\\d{1,3}\\\\.\\\\d{1,3}\\\\.\\\\d{1,3})((?=[^A-Za-z0-9])|$)", "mask_with": "IP"},\n          {"regex_pattern":"((?<=[^A-Za-z0-9])|^)([\\\\-\\\\+]?\\\\d+)((?=[^A-Za-z0-9])|$)", "mask_with": "NUM"},\n          ]\n    ]\n```\n\n## Persistence\n\nThe persistence feature saves and loads a snapshot of Drain3 state in a (compressed) json format. This feature adds\nrestart resiliency to Drain allowing continuation of activity and maintain learned knowledge across restarts.\n\nDrain3 state includes the search tree and all the clusters that were identified up until snapshot time.\n\nThe snapshot also persist number of log messages matched each cluster, and it\'s `cluster_id`.\n\nAn example of a snapshot:\n\n```json\n{\n  "clusters": [\n    {\n      "cluster_id": 1,\n      "log_template_tokens": [\n        "aa",\n        "aa",\n        "<*>"\n      ],\n      "py/object": "drain3_core.LogCluster",\n      "size": 2\n    },\n    {\n      "cluster_id": 2,\n      "log_template_tokens": [\n        "My",\n        "IP",\n        "is",\n        "<IP>"\n      ],\n      "py/object": "drain3_core.LogCluster",\n      "size": 1\n    }\n  ]\n}\n```\n\nThis example snapshot persist two clusters with the templates:\n\n`["aa", "aa", "<*>"]` - occurs twice\n\n`["My", "IP", "is", "<IP>"]` - occurs once\n\nSnapshots are created in the following events:\n\n- `cluster_created` - in any new template\n- `cluster_template_changed` - in any update of a template\n- `periodic` - after n minutes from the last snapshot. This is intended to save cluster sizes even if no new template\n  was identified.\n\nDrain3 currently supports the following persistence modes:\n\n- **Kafka** - The snapshot is saved in a dedicated topic used only for snapshots - the last message in this topic is the\n  last snapshot that will be loaded after restart. For Kafka persistence, you need to provide: `topic_name`. You may\n  also provide other `kwargs` that are supported by `kafka.KafkaConsumer` and `kafka.Producer` e.g `bootstrap_servers`\n  to change Kafka endpoint (default is `localhost:9092`).\n\n- **Redis** - The snapshot is saved to a key in Redis database (contributed by @matabares).\n\n- **File** - The snapshot is saved to a file.\n\n- **Memory** - The snapshot is saved an in-memory object.\n\n- **None** - No persistence.\n\nDrain3 persistence modes can be easily extended to another medium / database by inheriting\nthe [PersistenceHandler](drain3/persistence_handler.py) class.\n\n## Training vs. Inference modes\n\nIn some use-cases, it is required to separate training and inference phases.\n\nIn training phase you should call `template_miner.add_log_message(log_line)`. This will match log line against an\nexisting cluster (if similarity is above threshold) or create a new cluster. It may also change the template of an\nexisting cluster.\n\nIn inference mode you should call `template_miner.match(log_line)`. This will match log line against previously learned\nclusters only. No new clusters are created and templates of existing clusters are not changed. Match to existing cluster\nhas to be perfect, otherwise `None` is returned. You can use persistence option to load previously trained clusters\nbefore inference.\n\n## Memory efficiency\n\nThis feature limits the max memory used by the model. It is particularly important for large and possibly unbounded log\nstreams. This feature is controlled by the `max_clusters\u200b` parameter, which sets the max number of clusters/templates\ntrarcked by the model. When the limit is reached, new templates start to replace the old ones according to the Least\nRecently Used (LRU) eviction policy. This makes the model adapt quickly to the most recent templates in the log stream.\n\n## Parameter Extraction\n\nDrain3 supports retrieving an ordered list of variables in a log message, after its template was mined. Each parameter\nis accompanied by the name of the mask that was matched, or `*` for the catch-all mask.\n\nParameter extraction is performed by generating a regular expression that matches the template and then applying it on\nthe log message. When `exact_matching` is enabled (by default), the generated regex included the regular expression\ndefined in relevant masking instructions. If there are multiple masking instructions with the same name, either match\ncan satisfy the regex. It is possible to disable exact matching so that every variable is matched against a\nnon-whitespace character sequence. This may improve performance on expanse of accuracy.\n\nParameter extraction regexes generated per template are cached by default, to improve performance. You can control cache\nsize with the ` MASKING/parameter_extraction_cache_capacity` configuration parameter.\n\nSample usage:\n\n```python\nresult = template_miner.add_log_message(log_line)\nparams = template_miner.extract_parameters(\n    result["template_mined"], log_line, exact_matching=True)\n```\n\nFor the input `"user johndoe logged in 11 minuts ago"`, the template would be:\n\n```\n"user <:*:> logged in <:NUM:> minuts ago"\n```\n\n... and the extracted parameters:\n\n```\n[\n  ExtractedParameter(value=\'johndoe\', mask_name=\'*\'), \n  ExtractedParameter(value=\'11\', mask_name=\'NUM\')\n]\n```\n\n## Installation\n\nDrain3 is available from [PyPI](https://pypi.org/project/drain3). To install use `pip`:\n\n```\npip3 install drain3\n```\n\nNote: If you decide to use Kafka or Redis persistence, you should install relevant client library explicitly, since it\nis declared as an extra (optional) dependency, by either:\n\n```\npip3 install kafka-python\n```\n\n-- or --\n\n```\npip3 install redis\n```\n\n## Examples\n\nIn order to run the examples directly from the repository, you need to install dependencies. You can do that using *\npipenv* by executing the following command (assuming pipenv already installed):\n\n```shell\npython3 -m pipenv sync\n```\n\n#### Example 1 - `drain_stdin_demo`\n\nRun [examples/drain_stdin_demo.py](examples/drain_stdin_demo.py) from the root folder of the repository by:\n\n```\npython3 -m pipenv run python -m examples.drain_stdin_demo\n```\n\nThis example uses Drain3 on input from stdin and persist to either Kafka / file / no persistence.\n\nChange `persistence_type` variable in the example to change persistence mode.\n\nEnter several log lines using the command line. Press `q` to end online learn-and-match mode.\n\nNext, demo goes to match (inference) only mode, in which no new clusters are trained and input is matched against\npreviously trained clusters only. Press `q` again to finish execution.\n\n#### Example 2 - `drain_bigfile_demo`\n\nRun [examples/drain_bigfile_demo](examples/drain_bigfile_demo.py) from the root folder of the repository by:\n\n```\npython3 -m pipenv run python -m examples.drain_bigfile_demo\n```\n\nThis example downloads a real-world log file (of an SSH server) and process all lines, then prints result clusters,\nprefix tree and performance statistics.\n\n#### Sample config file\n\nAn example `drain3.ini` file with masking instructions can be found in the [examples](examples) folder as well.\n\n## Contributing\n\nOur project welcomes external contributions. Please refer to [CONTRIBUTING.md](CONTRIBUTING.md) for further details.\n\n## Change Log\n\n##### v0.9.11\n\n* Fixed possible DivideByZero error when the profiler is enabled - [Issue #65](https://github.com/IBM/Drain3/issues/65). \n\n##### v0.9.10\n\n* Fixed compatibility issue with Python 3.10 caused by removal of `KeysView`.\n\n##### v0.9.9\n\n* Added support for accurate log message parameter extraction in a new function - `extract_parameters()`. The\n  function `get_parameter_list()` is deprecated (Thanks to *@Impelon*).\n* Refactored `AbstractMaskingInstruction` as a base class for `RegexMaskingInstruction`, allowing to introduce other\n  types of masking mechanisms.\n\n##### v0.9.8\n\n* Added an option `full_search_strategy` option in `TemplateMiner.match()` and `Drain.match()`. See more info at\n  Issue [#48](https://github.com/IBM/Drain3/issues/48).\n* Added an option to disable parameterization of tokens that contains digits in\n  configuration: `TemplateMinerConfig.parametrize_numeric_tokens`\n* Loading Drain snapshot now only restores clusters state and not configuration parameters. This improves backwards\n  compatibility when introducing new Drain configuration parameters.\n\n##### v0.9.7\n\n* Fixed bug in original Drain: log clusters were created multiple times for log messages with fewer tokens\n  than `max_node_depth`.\n* Changed `depth` property name to a more descriptive name `max_node_depth` as Drain always subtracts 2 of `depth`\n  argument value. Also added `log_cluster_depth` property to reflect original value of depth argument (Breaking Change).\n* Restricted `depth` param to minimum sensible value of 3.\n* Added log cluster count to nodes in `Drain.print_tree()`\n* Added optional log cluster details to `Drain.print_tree()`\n\n##### v0.9.6\n\n* Fix issue https://github.com/IBM/Drain3/issues/38: Unnecessary update of LRU cache in case `max_clusters` is used (\n  thanks *@StanislawSwierc*).\n\n##### v0.9.5\n\n* Added: `TemplateMiner.match()` function for fast matching against existing clusters only.\n\n##### v0.9.4\n\n* Added: `TemplateMiner.get_parameter_list()` function to extract template parameters for raw log message (thanks to *\n  @cwyalpha*)\n* Added option to customize mask wrapper - Instead of the default `<*>`, `<NUM>` etc, you can select any wrapper prefix\n  or suffix by overriding `TemplateMinerConfig.mask_prefix` and `TemplateMinerConfig.mask_prefix`\n* Fixed: config `.ini` file is always read from same folder as source file in demos in tests (thanks *@RobinMaas95*)\n\n##### v0.9.3\n\n* Fixed: comparison of type int with type str in function `add_seq_to_prefix_tree` #28 (bug introduced at v0.9.1)\n\n##### v0.9.2\n\n* Updated jsonpickle version\n* Keys `id_to_cluster` dict are now persisted by jsonpickle as `int` instead of `str` to avoid keys type conversion on\n  load snapshot which caused some issues.\n* Added cachetools dependency to `setup.py`.\n\n##### v0.9.1\n\n* Added option to configure `TemplateMiner` using a configuration object (without `.ini` file).\n* Support for `print_tree()` to a file/stream.\n* Added `MemoryBufferPersistence`\n* Added unit tests for state save/load.\n* Bug fix: missing type-conversion in state loading, introduced in v0.9.0\n* Refactor: Drain prefix tree keys are now of type `str` also for 1st level\n  (was `int` before), for type consistency.\n\n##### v0.9.0\n\n* Decrease memory footprint of the main data structures.\n* Added `max_clusters` option to limit the number of tracked clusters.\n* Changed cluster identifier type from str to int\n* Added more unit tests and CI\n\n##### v0.8.6\n\n* Added `extra_delimiters` configuration option to Drain\n\n##### v0.8.5\n\n* Profiler improvements\n\n##### v0.8.4\n\n* Masking speed improvement\n\n##### v0.8.3\n\n* Fix: profiler state after load from snapshot\n\n##### v0.8.2\n\n* Fixed snapshot backward compatibility to v0.7.9\n\n##### v0.8.1\n\n* Bugfix in profiling configuration read\n\n##### v0.8.0\n\n* Added time profiling support (disabled by default)\n* Added cluster ID to snapshot reason log (credit: @boernd)\n* Minor Readability and documentation improvements in Drain\n\n##### v0.7.9\n\n* Fix: `KafkaPersistence` now accepts also `bootstrap_servers` as kwargs.\n\n##### v0.7.8\n\n* Using `kafka-python` package instead of `kafka` (newer).\n* Added support for specifying additional configuration as `kwargs` in Kafka persistence handler.\n\n##### v0.7.7\n\n* Corrected default Drain config values.\n\n##### v0.7.6\n\n* Improvement in config file handling (Note: new sections were added instead of `DEFAULT` section)\n\n##### v0.7.5\n\n* Made Kafka and Redis optional requirements\n \n',
    'author': 'IBM Research Haifa',
    'author_email': 'david.ohana@ibm.com',
    'maintainer': 'Yihao Chen(Superskyyy)',
    'maintainer_email': 'yihaochen@apache.org',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: MIT
# This file implements the Drain algorithm for log parsing.
# Based on https://github.com/logpai/logparser/blob/master/logparser/Drain/Drain.py by LogPAI team

from drain3.drain import DrainBase, LogCluster, Node


class JaccardDrain(DrainBase):
    """
    add a new matching pattern to the log cluster.
    Cancels log message length as  first token.
    Drain that uses Jaccard similarity to match log messages.
    """

    def tree_search(self, root_node: Node, tokens: list, sim_th: float, include_params: bool):
        # at first level, children are grouped by token (The first word in tokens)
        token_count = len(tokens)
        # cur_node = root_node.key_to_child_node.get(str(token_count))

        if not tokens:
            token_first = ""
            cur_node = root_node.key_to_child_node.get(token_first)
        else:
            token_first = tokens[0]
            cur_node = root_node.key_to_child_node.get(token_first)

        # no template with same token count yet
        if cur_node is None:
            return None

        # handle case of empty log string - return the single cluster in that group
        if token_count == 0:
            return self.id_to_cluster.get(cur_node.cluster_ids[0])

        # find the leaf node for this log - a path of nodes matching the first N tokens (N=tree depth)
        cur_node_depth = 1  # first level is 1 <root>

        for token in tokens[1:]:
            # at max depth
            if cur_node_depth >= self.max_node_depth:
                break

            # this is last token
            # It starts with the second word, so the sentence length -1
            if cur_node_depth == token_count - 1:
                break

            key_to_child_node = cur_node.key_to_child_node
            cur_node = key_to_child_node.get(token)

            if cur_node is None:  # no exact next token exist, try wildcard node
                cur_node = key_to_child_node.get(self.param_str)
            if cur_node is None:  # no wildcard node exist
                return None

            cur_node_depth += 1

        # get best match among all clusters with same prefix, or None if no match is above sim_th
        cluster = self.fast_match(cur_node.cluster_ids, tokens, sim_th, include_params)

        return cluster

    def add_seq_to_prefix_tree(self, root_node, cluster: LogCluster):
        token_count = len(cluster.log_template_tokens)
        # Determine if the string is empty
        if not cluster.log_template_tokens:
            token_first = ""
        else:
            token_first = cluster.log_template_tokens[0]
        if token_first not in root_node.key_to_child_node:
            first_layer_node = Node()
            root_node.key_to_child_node[token_first] = first_layer_node
        else:
            first_layer_node = root_node.key_to_child_node[token_first]

        cur_node = first_layer_node

        # handle case of empty log string
        if token_count == 0:
            cur_node.cluster_ids = [cluster.cluster_id]
            return

        # test_add_shorter_than_depth_message : only one word add into current node
        if token_count == 1:
            # clean up stale clusters before adding a new one.
            new_cluster_ids = []
            for cluster_id in cur_node.cluster_ids:
                if cluster_id in self.id_to_cluster:
                    new_cluster_ids.append(cluster_id)
            new_cluster_ids.append(cluster.cluster_id)
            cur_node.cluster_ids = new_cluster_ids

        current_depth = 1
        for token in cluster.log_template_tokens[1:]:
            # if at max depth or this is last token in template - add current log cluster to the leaf node
            # It starts with the second word, so the sentence length -1
            if current_depth >= self.max_node_depth or current_depth >= token_count - 1:
                # clean up stale clusters before adding a new one.
                new_cluster_ids = []
                for cluster_id in cur_node.cluster_ids:
                    if cluster_id in self.id_to_cluster:
                        new_cluster_ids.append(cluster_id)
                new_cluster_ids.append(cluster.cluster_id)
                cur_node.cluster_ids = new_cluster_ids
                break

            # if token not matched in this layer of existing tree.
            if token not in cur_node.key_to_child_node:
                if self.parametrize_numeric_tokens and self.has_numbers(token):
                    if self.param_str not in cur_node.key_to_child_node:
                        new_node = Node()
                        cur_node.key_to_child_node[self.param_str] = new_node
                        cur_node = new_node
                    else:
                        cur_node = cur_node.key_to_child_node[self.param_str]

                else:
                    if self.param_str in cur_node.key_to_child_node:
                        if len(cur_node.key_to_child_node) < self.max_children:
                            new_node = Node()
                            cur_node.key_to_child_node[token] = new_node
                            cur_node = new_node
                        else:
                            cur_node = cur_node.key_to_child_node[self.param_str]
                    else:
                        if len(cur_node.key_to_child_node) + 1 < self.max_children:
                            new_node = Node()
                            cur_node.key_to_child_node[token] = new_node
                            cur_node = new_node
                        elif len(cur_node.key_to_child_node) + 1 == self.max_children:
                            new_node = Node()
                            cur_node.key_to_child_node[self.param_str] = new_node
                            cur_node = new_node
                        else:
                            cur_node = cur_node.key_to_child_node[self.param_str]

            # if the token is matched
            else:
                cur_node = cur_node.key_to_child_node[token]

            current_depth += 1

    # seq1 is a template, seq2 is the log to match
    def get_seq_distance(self, seq1: tuple, seq2: list, include_params: bool):
        # Jaccard index, It is used to measure the similarity of two sets.
        # The closer its value is to 1, the more common members the two sets have, and the higher the similarity.

        # sequences are empty - full match
        if len(seq1) == 0:
            return 1.0, 0

        param_count = 0

        for token1 in seq1:
            if token1 == self.param_str:
                param_count += 1

        # If the token and the data have the same length, and there are param_str in the token
        if len(seq1) == len(seq2) and param_count > 0:
            # seq2 removes the param_str position
            seq2 = [x for i, x in enumerate(seq2) if seq1[i] != self.param_str]

        # If there are param_str, they are removed from the coefficient calculation
        if include_params:
            seq1 = [x for x in seq1 if x != self.param_str]

        # Calculate the Jaccard coefficient
        ret_val = len(set(seq1) & set(seq2)) / len(set(seq1) | set(seq2))

        # Jaccard coefficient calculated under the same conditions has a low simSep value
        # So gain is applied to the calculated value (The test case test_add_log_message_sim_75)
        ret_val = ret_val * 1.3 if ret_val * 1.3 < 1 else 1

        return ret_val, param_count

    # seq1:tonkens->list seq2:template->tuple
    def create_template(self, seq1: list, seq2: tuple):

        inter_set = set(seq1) & set(seq2)

        # test_max_clusters_lru_multiple_leaf_nodes
        # Update param_str at different positions with the same length
        if len(seq1) == len(seq2):
            ret_val = list(seq2)
            for i, (token1, token2) in enumerate(zip(seq1, seq2)):
                if token1 != token2:
                    ret_val[i] = self.param_str
        # param_str is updated at the new position with different length
        else:
            # Take the template with long length
            ret_val = seq1 if len(seq1) > len(seq2) else list(seq2)
            for i, token in enumerate(ret_val):
                if token not in inter_set:
                    ret_val[i] = self.param_str

        return ret_val

    def match(self, content: str, full_search_strategy="never"):

        assert full_search_strategy in ["always", "never", "fallback"]

        # Because the template length and data are not equal in length, Jaccard distance required_sim_th != 1
        required_sim_th = 0.8
        content_tokens = self.get_content_as_tokens(content)

        def full_search():
            all_ids = self.get_clusters_ids_for_seq_len(content_tokens[0])
            cluster = self.fast_match(all_ids, content_tokens, required_sim_th, include_params=True)
            return cluster

        if full_search_strategy == "always":
            return full_search()

        match_cluster = self.tree_search(self.root_node, content_tokens, required_sim_th, include_params=True)
        if match_cluster is not None:
            return match_cluster

        if full_search_strategy == "never":
            return None

        return full_search()

